#pragma TextEncoding = "UTF-8"
#pragma rtGlobals=3
#pragma DefaultTab={3,20,4}
#pragma ModuleName=ColorByIndex
#pragma IgorVersion=8
#pragma version=1.40

// Project Updater header
// http://www.igorexchange.com/project/Updater
static constant kProjectID=22305 // the project node on IgorExchange
static strconstant ksShortTitle="ColorByIndex" // the project short title on IgorExchange

// Set the color of each trace based on a selectable color table and index wave.
// Alternatively, the index value for each trace can be edited in the listbox control.
// A color scale is added to the graph window when the traces are recoloured.

// If a custom color table is selected, when traces are recolored a color
// table wave will be created in the current data folder with a name
// formed by appending "ColorTable" to the graph name. Any existing wave
// with this name will be overwritten.

// Tip: hidden traces do not appear in the listbox. Drag tracenames to reorder.

// prefs header
static strconstant ksPackageName = "ColorByIndex"
static strconstant ksPrefsFileName = "acwColorByIndex.bin"
static constant kPrefsVersion = 100

// 200 byte structure
static structure PackagePrefs
	uint16 version	// Preferences structure version number. 100 means 1.00.
	uint16 options // bit 0, show CT; bit 1, live
	char tablenum
	char reserved[200 - 2*2 - 1]	// Reserved for future use
endstructure

static function PrefsSetDefaults(STRUCT PackagePrefs &prefs)
	prefs.version  = kPrefsVersion
	prefs.options  = 1
	prefs.tablenum = IgorVersion()>=9 ? 61 : 2
	SavePackagePreferences ksPackageName, ksPrefsFileName, 0, prefs
	return 0
end

static function LoadPrefs(STRUCT PackagePrefs &prefs)
	LoadPackagePreferences ksPackageName, ksPrefsFileName, 0, prefs
	if (V_flag!=0 || V_bytesRead==0 || prefs.version!=kPrefsVersion)
		PrefsSetDefaults(prefs)
	endif
end

menu "Graph"
	"Color Traces by Index", /Q,  ColorByIndex#MakePanel()
end

// how to use this entry point function for external function calls

// Valid command (cmd) values are "Kill", "Init", and "Set"

// Commands may be combined with optional parameters:
// Use apply=1 to reset trace colors and update the color scale in the
//   graph window when live updating is turned off
// "Init" may be accompanied by an optional string parameter
//   Graph="GraphName" to select the target graph window, default = top graph
// ColorTable="ColorTableName" selects a standard color table, default is last-used color table
// CTwave=w1 selects a custom color table wave w1
// optional variable parameters vMin and vMax set the start and end index values for the color table
// order=mode selects an indexing scheme. Valid values for mode are "linear", "parabolic", "log", "exp",
//   "start color", "end color", and "clear"; default is "linear"
// IndexWave=w1 takes index values from wave w1
// reverseindex=value reverses the index values when value is non-zero; default = 0
// scale=value appends a color scale to the graph when value is non-zero; default = last used value
// live=0 and live=1 toggle live updating; default = last used value
// infomenu=0 combined with "Init" suppresses the blue info icon and the associated menu; default = 0

// examples
// ColorByIndex#do_cmd("Init", colortable="Grays", live=1)
// ColorByIndex#do_cmd("Set", IndexWave=root:foo, scale=0, apply=1)
static function do_cmd(string cmd,
	[string Graph, int apply, string ColorTable, wave/Z CTwave, variable vMin,
	 variable vMax, string order, wave/Z IndexWave, int reverseindex, int scale,
	 int live, int infomenu])
	
	variable vrtn = 0
	
	strswitch(cmd)
		case "Kill":
		case "Kill Panel":
			KillWindow/Z CTpanel
			return WinType("CTpanel") == 7
		case "Append Panel":
		case "Init":			
			infomenu = paramisdefault(infomenu) ? 1 : infomenu			
			if (ParamIsDefault(Graph))
				vrtn = MakePanel(infomenu=infomenu)
			else
				vrtn = MakePanel(Graph=Graph, infomenu=infomenu)
			endif
			if (!vrtn)
				return vrtn
			endif
			break
		case "Set":
			break
		default:
			return 0
	endswitch
		
	DFREF dfr = getDFR()
	
	if (!ParamIsDefault(ColorTable))
		int mode = WhichListItem(ColorTable, CTabList(), ";", 0, 0) + 1
		if (mode > 0)
			PopupMenu popCT	win=CTpanel, mode=mode
		endif
		CheckBox chkCustom win=CTpanel, value=0
		UseCustomCT(0)
	endif
	
	if (!ParamIsDefault(CTwave))
		if (!WaveExists(CTwave))
			return 0
		endif
		PopupMenu popCustom	win=CTpanel, popmatch=NameOfWave(CTwave)
		UseCustomCT(1)
	endif
	
	if (!ParamIsDefault(vMin))
		NVAR/Z/SDFR=dfr gMin = vMin
		if (NVAR_Exists(gMin))
			gMin = vMin
		endif
	endif
	
	if (!ParamIsDefault(vMax))
		NVAR/Z/SDFR=dfr gMax = vMax
		if (NVAR_Exists(gMax))
			gMax = vMax
		endif
	endif
	
	if (!ParamIsDefault(reverseindex))
		ControlInfo/W=CTpanel chkReverse
		if (reverseindex != v_value)
			ReverseIndices()
		endif
		CheckBox chkReverse win=CTpanel, value=reverseindex
	endif
	
	if (!ParamIsDefault(order))
		if (ResetIndices(order))
			PopupMenu popSetIndices win=CTpanel, popmatch=order, userdata=order
		endif
	endif
	
	if (!ParamIsDefault(IndexWave))
		if (!WaveExists(IndexWave))
			return 0
		endif
		if (ValuesFromWave(indexwave=IndexWave))
			PopupMenu popSetIndices win=CTpanel, popmatch="from wave"
		endif
	endif
	
	if (!ParamIsDefault(scale))
		CheckBox chkScale win=CTpanel, value=scale
	endif
	
	if (!ParamIsDefault(live))
		CheckBox chkLive win=CTpanel, value=live
		Button btnDoIt win=CTpanel, disable=live
	endif
	
	apply = ParamIsDefault(apply) ? 0 : apply
	vrtn = RecolourTraces(apply)
	
	return vrtn
end

static function/S GetGraph()
	GetWindow CTpanel activeSW
	return ParseFilePath(0, S_Value, "#", 0, 0)
end

static function/DF getDFR([string package])
	package = SelectString(ParamIsDefault(package), package, "ColorByIndex")
	DFREF dfr = root:Packages:$package
	if (DataFolderRefStatus(dfr) != 1)
		NewDataFolder/O root:Packages
		NewDataFolder/O root:Packages:$package
	endif
	DFREF dfr = root:Packages:$package
	return dfr
end

static function HookCTtraces(STRUCT WMWinHookStruct &s)
	string strGraph = ParseFilePath(0, s.winName, "#", 0, 0)
	if (stringmatch(s.winName, "*CTpanel"))
		if (s.eventCode == 2) // kill
			
			STRUCT PackagePrefs prefs
			LoadPrefs(prefs)
			ControlInfo/W=CTpanel popCT
			prefs.tablenum = V_Value - 1
			int options = 0
			ControlInfo/W=CTpanel chkScale
			options += 1 * v_value
			ControlInfo/W=CTpanel chkLive
			options += 2 * v_value
			prefs.options = options
			SavePackagePreferences ksPackageName, ksPrefsFileName, 0, prefs
			
			SetWindow $strGraph hook(hCTtraces)=$""
			KillWindow/Z $s.winName#G0
			KillDataFolder/Z root:Packages:ColorByIndex
		endif
	else // graph window
		if (s.eventCode==8 && cmpstr(strGraph, s.winName)==0) // modified
			CheckListboxWaves(0, s.winName)
		endif
	endif
	return 0
end

static function MakePanel([string Graph, int infomenu])
	KillWindow/Z CTpanel
	DoUpdate
	
	if (!ParamIsDefault(Graph) && !strlen(Graph))
		Graph = WinName(0, 1)
	endif
	string strGraph = SelectString(ParamIsDefault(Graph), Graph, WinName(0, 1))
	if (WinType(strGraph) != 1)
		DoAlert 0, "No graph window"
		return 0
	endif
	
	infomenu = ParamIsDefault(infomenu) ? 1 : infomenu
	
	STRUCT PackagePrefs prefs
	LoadPrefs(prefs)
	int showCT = prefs.options&1
	int live = (prefs.options&2) != 0
	int tablenum = prefs.tablenum
	tablenum = limit(tablenum, 0, ItemsInList(CTabList()) - 1)
	DFREF dfr = getDFR()
	
	NVAR/Z/SDFR=dfr vMin, vMax
	if (NVAR_Exists(vMin)==0 || NVAR_Exists(vMax)==0)
		variable/G dfr:vMin = 0, dfr:vMax = 100
		NVAR/Z/SDFR=dfr vMin, vMax
		vMax = 10 * (ItemsInList(TraceNameList(strGraph, ";", 5))-1)
		vMax = vMax <= 0 ? 100 : vMax
	endif
	
	Make/T/O/N=(0,2) dfr:CTtraces/wave=CTtraces, dfr:wHelp/wave=wHelp
	Make/O/N=(0,2) dfr:selwave/wave=selwave
	Make/T/O dfr:wTitle = {"trace (click and drag to reorder)", "index"}
	
	variable x0 = 10
	
	NewPanel/HOST=$strGraph/EXT=0/K=1/N=CTpanel/W=(0,0,220,370) as "Color Traces by Index"
	
	SetWindow CTpanel userdata(version) = num2str(GetThisVersion())
	
	if (infomenu)
		Button btnInfo, pos={210,5}, align=1, size={15,15}, Proc=ColorByIndex#ButtonProc, title=""
		Button btnInfo, help={"Show Info"}, Picture=ColorByIndex#pInfo
	else
		TitleBox titleVersion pos={210,5}, align=1, title="Version = " + num2str(GetThisVersion()/100), frame=0, fsize=9
	endif
	
	GroupBox grpCT, pos={-8,10}, size={250,95}, title="Color table", fsize=12
	PopupMenu popCT, pos={x0,30}, size={200,20}, mode=tablenum+1, value=#"\"*COLORTABLEPOP*\""
	PopupMenu popCT, Proc=ColorByIndex#PopupProc, help={"Select one of Igor's built-in color tables"}
	CheckBox chkCustom, pos={x0,55}, title="CT wave", value=0, fsize=12, Proc=ColorByIndex#CheckBoxProc
	CheckBox chkCustom, help={"Use a custom a color table wave"}
	PopupMenu popCustom, pos={x0+70,55}, size={150,20}, value=ColorByIndex#ListOfColorTables()
	PopupMenu popCustom, Proc=ColorByIndex#PopupProc, disable=2, help={"Select or load a custom a color table wave"}
	SetVariable svMin, pos={x0,80}, size={132,14}, title="Table start index", fsize=12, Proc=ColorByIndex#SetVariableProc
	SetVariable svMin, limits={-Inf,Inf,0}, value=dfr:vMin, help={"Index value for color table minimum"}
	SetVariable svMax, pos={x0+136,80}, size={65,14}, title="End", fsize=12, Proc=ColorByIndex#SetVariableProc
	SetVariable svMax, limits={-Inf,Inf,0}, value=dfr:vMax, help={"Index value for color table maximum"}
	ListBox list0, pos={x0,110}, size={200,155}, listWave=CTtraces, selwave=selwave, userColumnResize=1
	ListBox list0, widths={5,1}, Proc=ColorByIndex#ListboxProc, titleWave=dfr:wTitle, helpWave=wHelp, mode=5
	
	GroupBox grpIndices, pos={-8,270}, size={240,45}, title="Set trace index values", fsize=12
	PopupMenu popSetIndices, pos={x0+10,290}, size={90,20}, fSize=12, Proc=ColorByIndex#PopupProc, title="", userdata(oldstring)="linear"
	PopupMenu popSetIndices value="linear;parabolic;log;exp;start color;end color;clear;\\M1-;from wave;", help={"Choose a method to set index values"}
	CheckBox chkReverse, pos={x0+120,291}, title="Reverse", value=0, fsize=12, Proc=ColorByIndex#CheckBoxProc, help={"Reverse the order of the index values"}
	CheckBox chkScale, pos={x0,320}, title="Display color scale", value=showCT, fsize=12, Proc=ColorByIndex#CheckBoxProc, help={"Show color scale on graph window\rColor scale can be edited to reverse axis, change display range, etc."}
	CheckBox chkLive, pos={x0,340}, title="Live graph updates", value=live, fsize=12, Proc=ColorByIndex#CheckBoxProc, help={"Changing index values and settings updates the trace colors immediately"}
	Button btnDoIt, pos={x0+150,340}, size={50,20}, title="Do It", Proc=ColorByIndex#ButtonProc
	Button btnDoIt, help={"Reset trace colors and update color scale"}, disable=live
	
	CheckListboxWaves(1, strGraph)
	SetWindow $(strGraph)#CTpanel hook(hCTtraces)=ColorByIndex#HookCTtraces
	SetWindow $strGraph hook(hCTtraces)=ColorByIndex#HookCTtraces
	
	RecolourTraces(0)
	return 1
end

static function InfoMenu()
	string cmd = "(Version: " + num2str(GetThisVersion()/100) + ";"
	cmd += "Visit web page;Email the developer;"
	PopupContextualMenu cmd
	if (V_flag == 2)
		BrowseURL /Z "https://www.wavemetrics.com/node/" + num2str(kProjectID)
	elseif (V_flag == 3)
		Make/T/free/N=5 wt = "mail"
		string newline = "\r\n"
		wt[3] = "subject=" + URLEncode(ksShortTitle + " Package for Igor Pro")
		wt[4] = "Type your message here" + newline + newline
		wt[4] += "VERSION:" + num2str(GetThisVersion()/100) + newline
		wt[4] += ReplaceString(";", IgorInfo(3), newline) + newline
		wt[4] = "body=" + URLEncode(wt[4])
		wt[1] = "tony withers"
		wt[2] = "uni-bayreuth de"
		sprintf cmd, "%sto:%s%%%d%s?%s&%s" wt[0], wt[1], 40, wt[2], wt[3], wt[4]
		BrowseURL/Z ReplaceString(" ", cmd, ".")
	endif
end


// ensure that tracenames in listbox match the plotted traces
static function CheckListboxWaves(int reset, string strGraph)
	
	DFREF dfr = getDFR()
	wave/T/SDFR=dfr CTtraces, wHelp
	wave/SDFR=dfr selwave
	NVAR/Z/SDFR=dfr vMin, vMax
	
	string strTraces = TraceNameList(strGraph, ";", 5)
	int numTraces = ItemsInList(strTraces)
	int numRows = DimSize(CTtraces, 0)
	int i, item, j
	for (i=numrows-1;i>=0;i--)
		if (!strlen(CTtraces[i][0]) || !WaveExists(TraceNameToWaveRef(strGraph, CTtraces[i][0])))
			DeletePoints/M=0 i, 1, CTtraces, selwave, wHelp
		else
			item = WhichListItem(CTtraces[i][0], strTraces)
			if (item > -1)
				strTraces = RemoveListItem(item, strTraces)
			endif
		endif
	endfor
	Redimension/N=(-1,2) CTtraces, selwave, wHelp
	
	numTraces = ItemsInList(strTraces)
	numRows = DimSize(CTtraces, 0)
	for (i=0;i<numTraces;i++)
		InsertPoints/M=0 numRows, 1, CTtraces, selwave, wHelp
		CTtraces[numRows][0] = StringFromList(i, strTraces)
		CTtraces[numRows][1] = ""
		numrows += 1
	endfor

	if (numRows)
		wHelp[][0] = CTtraces // in case we have long trace names
		selwave[][1] = 2 // make second column editable
		if (reset)
			CTtraces[][1] = num2str(vMin + p/(numRows-1)*(vMax-vMin))
		endif
	endif
	return 1
end

static function ResetIndices(string mode)
	DFREF dfr = getDFR()
	wave/T/SDFR=dfr CTtraces
	NVAR/Z/SDFR=dfr vMin, vMax
	
	variable numRows = DimSize(CTtraces, 0)
	if (numRows == 0)
		return 0
	endif
	variable range = vMax - vMin
	variable endP = numRows - 1
	ControlInfo/W=CTpanel chkReverse
	int rev = v_value
	
	strswitch(mode)
		case "clear":
			CTtraces[][1] = ""
			break
		case "linear":
			if (rev)
				CTtraces[][1] = num2str(round(vMin + (endP-p)/(endP) * range))
			else
				CTtraces[][1] = num2str(round(vMin + p/(endP) * range))
			endif
			break
		case "parabolic":
			if (rev)
				CTtraces[][1] = num2str(round(vMin + ((endP-p)/(endP))^2 * range))
			else
				CTtraces[][1] = num2str(round(vMin + (p/endP)^2 * range))
			endif
			break
		case "log":
			if (rev)
				CTtraces[][1] = num2str(round(vMin + log(endP-p+1)/log(numRows) * range))
			else
				CTtraces[][1] = num2str(round(vMin + log(p+1)/log(numRows) * range))
			endif
			break
		case "exp":
			if (rev)
				CTtraces[][1] = num2str(round(vMin + ((exp(endP-p)-1)/(exp(endP)-1)) * range))
			else
				CTtraces[][1] = num2str(round(vMin + ((exp(p)-1)/(exp(endP)-1)) * range))
			endif
			break
		case "start color":
			CTtraces[][1] = num2str(vMin)
			break
		case "end color":
			CTtraces[][1] = num2str(vMax)
			break
		case "from wave":
			return ValuesFromWave()
			break
		default:
			return 0
	endswitch
	return 1
end

static function CheckBoxProc(STRUCT WMCheckBoxAction &s) : CheckboxControl
		
	if (s.eventCode == 2)
		
		if (CheckUpdated(1))
			return 0
		endif
		
		strswitch (s.ctrlName)
			case "chkCustom" :
				UseCustomCT(s.checked)
				break
			case "chkReverse" :
				ReverseIndices()
				break
			case "chkLive" :
				Button btnDoIt win=CTpanel, disable=s.checked
				break
		endswitch
		RecolourTraces(0)
	endif
	return 0
end

static function UseCustomCT(int activate)
	CheckBox chkCustom win=CTpanel, value=activate
	PopupMenu popCT, win=CTpanel, disable = 2*activate
	PopupMenu popCustom, win=CTpanel, disable = 2*(!activate)
	KillWindow/Z CTpanel#G0
	if (activate)
		ControlInfo/W=CTpanel popCustom
		wave ctab = CreateCustomColorIndexWave(S_Value, 0)
		wave img = CreateImageWave(ctab)
		Display/W=(10,30,210,50)/HOST=CTpanel/N=G0
		AppendImage/T/W=CTpanel#G0 img
		ModifyGraph/W=CTpanel#G0 margin=-1, frameStyle=1, standoff=0, tick=3, noLabel=2
		SetAxis/A/W=CTpanel#G0
		SetWindow CTpanel#G0 activeChildFrame=0
		SetActiveSubwindow ##
	endif
end

static function ReverseIndices()
	wave/T/SDFR=getDFR() CTtraces
	Duplicate/free/T CTtraces wSorted
	Make/free/N=(DimSize(wSorted, 0)) wSortOrder = p
	SortColumns/R keyWaves=wSortOrder, sortWaves=wSorted
	CTtraces[][1] = wSorted
end

static function PopupProc(STRUCT WMPopupAction &s) : PopupMenuControl
	
	switch (s.eventCode)
		case 2:
			if (CheckUpdated(0))
				return 0
			endif
			
			if (cmpstr(s.ctrlName, "popCustom") == 0)
				string strCT
				if (cmpstr(s.popstr, "Load...") == 0)
					// no way to autoposition a help window :(
					DisplayHelpTopic "Color Table Waves in the Igor Pro Folder"
					PopupMenu $s.ctrlName, win=$s.win, popmatch="none"
					strCT = ""
				else
					strCT = s.popstr
				endif
				
				wave ctab = CreateCustomColorIndexWave(strCT, 0)
				CreateImageWave(ctab)
			endif
			
			if (cmpstr(s.ctrlName, "popSetIndices") == 0)
				string exitpop = s.popstr
				if (ResetIndices(s.popStr))
					PopupMenu $s.ctrlName win=$s.win, userdata(oldstring)=exitpop
				else
					PopupMenu $s.ctrlName win=$s.win, popmatch=GetUserData(s.win, s.ctrlName, "oldstring")
				endif
			endif
	
		case 3:
		case 4:
			// events 2, 3 or 4
			RecolourTraces(0)
			break
	endswitch
		
	return 0
end

static function ListboxProc(STRUCT WMListboxAction &s) : ListBoxControl
	
	switch (s.eventCode)
		case 1: // mousedown
			if (s.col==0)
				ListBox $s.ctrlName, win=$s.win, userdata(drag)=num2str(s.row)
				//userdata(drag) indicates dragging is active, cleared on mouseup
			endif
			break
		case 2: // mouseup
			ListBox $s.ctrlName, win=$s.win, userdata(drag)=""
			break
		case 4: // selection of s.row
			if (s.col != 0)
				return 0
			endif
			variable dragNum = str2num(GetUserData(s.win,s.ctrlName,"drag"))
			if (numtype(dragNum)!=0 || min(dragNum,s.row)<0 || max(dragNum,s.row)>=numpnts(s.listwave))
				return 0
			endif
			Make/N=(DimSize(s.selwave, 0))/free  order
			Duplicate/free/T s.listwave SortedListwave
			order = (p == dragNum) ? s.row - 0.5 + (s.row > dragNum) : p
			SortColumns keyWaves=order, sortWaves=SortedListwave
			s.listwave[][0] = SortedListwave
			ListBox $s.ctrlName, win=$s.win, userdata(drag)=num2str(s.row)
			RecolourTraces(0)
			break
		case 7: // end edit
			s.listWave[s.row][s.col] = ReplaceString("NaN", num2str(str2num(s.listWave[s.row][s.col])), "")
			RecolourTraces(0)
			break
	endswitch
	return 0
end

static function SetVariableProc(STRUCT WMSetVariableAction &s) : SetVariableControl
	switch (s.eventCode)
		case 2 :
		case 3 :
		case 4 :
		case 5 :
		case 8 :
			RecolourTraces(0)
			break
	endswitch
	return 0
end

static function ButtonProc(STRUCT WMButtonAction &s) : ButtonControl
		
	if (s.eventcode == 2)
		
		if (CheckUpdated(1))
			return 0
		endif
			
		strswitch(s.ctrlName)
			case "btnDoIt":
				RecolourTraces(1)
				break
			case "btnInfo" :
				InfoMenu()
				break
		endswitch
	endif
	return 0
end

static function/wave ChooseIndexWave()
	wave/T/SDFR=getDFR() CTtraces
	int rows = DimSize(CTtraces, 0)
	if (!rows)
		return $""
	endif
	
	string strOptions = ""
	sprintf strOptions, "DIMS:1,MINROWS:%d,MAXROWS:%d,TEXT:0", rows, rows
	
	string strList = WaveList("*", ";", strOptions)
	if (!strlen(strList))
		DoAlert 0, "Wave must have " + num2str(rows) + " points, one for each trace"
		return $""
	endif

	string strWave = ""
	Prompt strWave, "Select index wave", Popup, strList
	DoPrompt "Select wave containing index values", strWave
	if (v_flag)
		return $""
	endif
	return $strWave
end

static function ValuesFromWave([wave/Z indexwave])
	wave/T/SDFR=getDFR() CTtraces
	int rows = DimSize(CTtraces, 0)
	if (!rows)
		return 0
	endif
	
	if (ParamIsDefault(indexwave))
		wave/Z indexwave = ChooseIndexWave()
		if (!WaveExists(indexwave))
			return 0
		endif
	elseif (!WaveExists(indexwave) || numpnts(indexwave)!=rows)
		return 0
	endif
	
	ControlInfo/W=CTpanel chkReverse
	if (v_value)
		CTtraces[][1] = num2str(indexwave[rows - 1 - p])
	else
		CTtraces[][1] = num2str(indexwave[p])
	endif
	CTtraces[][1] = ReplaceString("NaN", CTtraces, "")

	return 1
end

static function RecolourTraces(int doit)
	int live = 0
	if (!doit)
		ControlInfo/W=CTpanel chkLive
		if (v_value == 0)
			return 0
		endif
		live = 1
	endif
	
	Button btnDoIt win=CTpanel, disable=live
	
	string strGraph = GetGraph()
	
	ControlInfo/W=CTpanel chkCustom
	int custom = v_value
	string strCT = ""
	
	int error = 0
	DFREF dfr = getDFR()
	NVAR/Z/SDFR=dfr vMin, vMax
	wave/T CTtraces = dfr:CTtraces
	
	if (custom)
		ControlInfo/W=CTpanel popCustom
		if (FindListItem(S_value, "Load...;none;") >=0)
			return 0
		endif
		string strName = CleanupName(strGraph + "ColorTable", 0)
		Duplicate/O dfr:M_colors $strName/wave = M_colors
	else
		ControlInfo/W=CTpanel popCT
		strCT = S_value
		DFREF dfSav = GetDataFolderDFR()
		SetDataFolder dfr
		ColorTab2Wave $strCT
		wave M_colors
		SetDataFolder dfSav
	endif
			
	variable numColors = DimSize(M_colors, 0)
	int i
	string strTrace
	variable pnt, val
	
	// colour the traces
	for (i=DimSize(CTtraces, 0)-1;i>=0;i--)
		strTrace = CTtraces[i][0]
		val = str2num(CTtraces[i][1])
		if (val != limit(val, min(vMin,vMax), max(vMin,vMax)))
			error = error | 1
		endif
		
		pnt = floor((str2num(CTtraces[i][1])-vMin)/(vMax-vMin)*(numColors))
		pnt = limit(pnt, 0, numColors-1)
		if (numtype(pnt))
			error = error | 2
			continue
		endif
		
		ModifyGraph /W=$strGraph rgb($strTrace) = (M_colors[pnt][0], M_colors[pnt][1], M_colors[pnt][2])
		
	endfor
	
	ControlInfo/W=$strGraph#CTpanel chkScale
	if (V_Value)
	
		// draw the colour scale
		if (custom)
			SetScale/I x, vMin, vMax, M_colors
			ColorScale/C/N=ct/F=0/A=RC cindex=M_colors
		else
			ColorScale/C/N=ct/F=0/A=RC ctab={vMin,vMax,$strCT,0}
		endif
	else
		ColorScale/K/N=ct
	endif
	
	if (error && !live)
		DoAlert 0, "Index of at least one trace is out of range.\rColors of out-of-range traces are not changed."
	endif
	
	return error == 0
end

// create an image wave to display color table in panel
static function/wave  CreateImageWave(wave ctab)
	DFREF dfr = getDFR()
	Make/O/N=(DimSize(ctab,0), 2, 3)/W/U dfr:CustomImg/wave=CustomImg
	CustomImg = ctab[p][r]
	return CustomImg
end

// create a color table index wave in the package folder
static function/wave CreateCustomColorIndexWave(string strCT, int reversed)
	DFREF dfr = getDFR()
	DFREF tableDFR = getDFR(package = "ColorTables")
	
	if (strlen(strCT))
		if (IgorVersion() >= 9)
			wave/Z w = FindWaveRecursive(strCT, tableDFR)
		else
			wave/Z w = $strCT
		endif
	endif
	if (!WaveExists(w))
		// make a dummy color table
		Make/O/U/I/N=(2,3) tableDFR:none = 60000
		wave w = tableDFR:none
	endif
	
	NVAR/Z/SDFR=dfr vMin, vMax
	Duplicate/O w dfr:M_colors/wave=M_colors
	SetScale/I x, vMin, vMax, M_colors

	return M_colors
end

static function/wave FindWaveRecursive(string strCT, DFREF dfr)
	int i
	for (i=0;1;i++)
		wave/Z w = WaveRefIndexedDFR(dfr, i)
		if (!WaveExists(w))
			break	// no more waves
		endif
		// see if it is our color table wave
		if (cmpstr(NameOfWave(w), strCT)==0)
			return w
		endif
	endfor
	
	// recurse into all subfolders
	variable numChildDataFolders = CountObjectsDFR(dfr, 4)
	for (i=0; i<numChildDataFolders; i+=1)
		DFREF childDFR = dfr:$GetIndexedObjNameDFR(dfr, 4, i)
		wave/Z w = FindWaveRecursive(strCT, childDFR)
		if (WaveExists(w))
			return w
		endif
	endfor
	
	return $""
end

// following three functions adapted from KBColorizeTraces.ipf
static function/S ListOfColorTables()
	if (IgorVersion() < 9)
		return RecurseColorTablesFolder("", GetDataFolderDFR(), recurse=0)
	endif
	string strColorTablesList = RecurseColorTablesFolder("", root:Packages:ColorTables)
	if (WhichListItem("none", strColorTablesList) < 0)
		strColorTablesList = "none;" + strColorTablesList
	endif
	return strColorTablesList + "Load...;"
end

// returns 1 if w is a color table wave (numeric, 3 or 4 columns, more than 1 row)
static function IsColorTableWave(WAVE w)

	if (WaveType(w, 1) != 1)
		return 0	// not numeric!
	endif
	
	int cols = DimSize(w, 1)
	if (cols < 3 || cols > 4)
		return 0
	endif
	
	if (DimSize(w, 0) < 2) // not enough rows
		return 0
	endif
	
	return 1
end

static function/S RecurseColorTablesFolder(string listSoFar, DFREF dfr, [int recurse])

	if (!DataFolderRefStatus(dfr))
		return listSoFar
	endif
	
	recurse = ParamIsDefault(recurse) ? 1 : recurse
	
	// iterate through waves in folder
	int i
	for (i=0;1;i++)
		wave/Z w = WaveRefIndexedDFR(dfr, i)
		if (!WaveExists(w))
			break // no more waves
		endif
		// see if it is a color table wave
		if (IsColorTableWave(w))
			listSoFar += NameOfWave(w) + ";"
		endif
	endfor
	
	// then recurse into all subfolders
	if (recurse)
		variable numChildDataFolders = CountObjectsDFR(dfr, 4)
		for (i=0; i<numChildDataFolders; i+=1)
			DFREF childDFR = dfr:$GetIndexedObjNameDFR(dfr, 4, i)
			listSoFar = RecurseColorTablesFolder(listSoFar, childDFR)
		endfor
	endif
	
	return listSoFar
end

// returns truth that this procedure file has been updated since initialisation
static function CheckUpdated(int restart)
	int version = GetThisVersion()
	int panelVersion = str2num(GetUserData("CTpanel", "", "version"))
	if (panelVersion != version)
		if (restart)
			DoAlert 0, "You have updated the package since this panel was created.\r\rThe package will restart to update the control panel."
			MakePanel()
		else
			DoAlert 0, "You have updated the package since this panel was created.\r\rPlease close and reopen the panel to continue."
		endif
		return 1
	endif
	return 0
end

// ProcedureVersion("") doesn't work for independent modules!
// extract procedure version from this file
static function GetThisVersion()

	variable procVersion
	// try the quick way
	#if exists("ProcedureVersion") == 3
	procVersion = ProcedureVersion("")
	if (procVersion)
		return 100 * procVersion
	endif
	#endif
	
	int maxLines = 20 // number of lines to search for version pragma
	int refNum, i
	string strHeader = ""
	string strLine = ""
	
	Open/R/Z refnum as FunctionPath("")
	if (refnum == 0)
		return 0
	endif
	for (i=0;i<maxLines;i+=1)
		FReadLine refNum, strLine
		strHeader += strLine
	endfor
	Close refnum
	wave/T ProcText = ListToTextWave(strHeader, "\r")
	Grep/Q/E="(?i)^#pragma[\s]*version[\s]*=" /LIST/Z ProcText
	if (v_flag != 0)
		return 0
	endif
	s_value = LowerStr(TrimString(s_value, 1))
	sscanf s_value, "#pragma version = %f", procVersion
	ProcVersion = (V_flag!=1 || ProcVersion<=0) ? 0 : ProcVersion

	return 100 * ProcVersion
end

// PNG: width= 90, height= 30
static Picture pInfo
	ASCII85Begin
	M,6r;%14!\!!!!.8Ou6I!!!"&!!!!?#R18/!3BT8GQ7^D&TgHDFAm*iFE_/6AH5;7DfQssEc39jTBQ
	=U#t@KS5u_NKc'hjQ<7_k>^N6#'iu's#@iL6?EN#,SR%ORqWY?g_DQ6ZakX-9m6UF`Y7B5's^*$E%B
	u3<(a:i;5<Ct\*^a)hT'K%0QC]d[0jPV;N@+9cG@fjB`p38^oR<*NngmMP4QXaLkr6j/Tp8k\$ZiA^
	F[J/@;&;!fa/0Y"8EKlO!*92tV8I!509L_n[prdb[0DD]b^77\4Qh+<X22WSL=^Tpl!>%R?bOV1l`F
	.H/kj.Q0YHO!p0Z_o!(t!"2!0>1<o&\&oX/k93#9I:/\ol*"]HFk2"&^kI,Op^Po$LJ:qE[=Zdn*%1
	/%c)$r:,4%r)#UgVCL9-4fb:)2uc00,SCF2g"FKhP:)ciSp&'<om1=hD5F)*"ddDT8%a$Z"blrKB1V
	t"bqBnEcNrlWWRU];Q?b;\6c%+`!0V&SGg^%NRV$e":,J(6+Q,'.qs8WXZ?]HQd?\@;-q_mAJ,]9,i
	ondb3ItHg(Zjc?GW[$h>t<@#!7L*gWDnag1eK/1,*>Cr1P3*D[BX`3`d.+2D9NUG(..rXX^7$1"@82
	!1KE^Ub4(b^NWMsF;Ht6&&;b$b#dun1L:o\#UFM.3gNK`n[l7Cp+c[X/cDaa?O<b.!b=-Pr2.^?1Nd
	)e2#U:Q=,SCE:&8^%*L$dX(UQj\B/Bea\(..rm00Qspp'-Bu;ci@[g2Ab(7h`\ZA=S45.(4!_jMmH)
	qUNK^MN[#3[D>08n\uMdd,!N;U*ru@<Dd(/5&X;O&;3@n7RotRD4?;%HAGk=G+fW.>tiHTqLp0W7=4
	1t\u)fWAJng`QIdQ"8TJr%$paau+6@''?G#H(RO^<AUs^^G[]\cSB@c5aI;=JigY9^3e#1T9SXPucg
	hq>mG`bOMiSe2FC2.Jjp@_T>`piIOWDnc;k%;&=gQn1M0p,[l:OIIe[!*"QkrNrEAXG*0N(r07]$fZ
	h/$5=fV3Tb,=$\psV/^EJ)4IFB]PEW'\qbg(SsMlA:JXaHWi@OR/f1s5k:MsaPoMc&J--#d&-rCGg1
	@0&:")(CjpXC94`bia)`#WJ(G?agO/a)!%!Za'1b>,JjlN'!=]&V5U;CI@Et9%o2-h+N"!1EAB5`,M
	Th&'lHW;8P&FJlPBUVCZ77C(9D;1qiXUW&&O5KD=a,V0/D%4B*&."qC)Xk?1%Zs:M/7mre"!%<ET2<
	bOD<cB%HGLYk/]p$:.Z,bn[7sj`?B9ZXXV6Rq]_Pti2gd\J3f(i:R&5_pm4KmDXm$aS/B^J!/^)#uG
	%6sfYult1cH[_1nRTfUip]H>d<$0>4r9NH8sWQ9)djkQAg3E#:]C\]o'uTN1rJBf7mE:cKPfJ<US@o
	sfC=1##U)g+pYUIm80gMI'&gcc'97IQ?8%%Ad+-[#D;.]AGGtPU1e&m.iUpnuW^=]Xm<&%d+jRgmR6
	q<!"K4)c%%V>s6.boU<k:W9@PN6o+0)@'G&kQ@q=8),*BN8CpR@_H@2O[8@=!9Ifs"n;A,h79HFe;'
	<AQ<<*SW3\4hYn'7c+!LCK<HgOtdE6\7T^B_C4`os.\g?L1-<OUon"8QFD=fRPcBo,tYs-%!5e<il-
	gK-hM%Q-;cW@=dI0o(g+V`I[7Hll^gmF1C8U`9;n:SDf?'4kXX)YA6CN6o]r]qT!s1+^3pK)B6pt_B
	a+O@`IRg8=:@XkJL"[Wg"k-TK%RqsK*5:)MTX360]U/4Wi@77;Jt3uP7<CJ.#Pn:q"VM]0-o&(8.o9
	`MEo2(5'MS@J+Rq@q<5kKI!A\?`c`\j5Xs.C`^\u^];Gr<!]0NL(T`^;0>7#+XNrGfbj/DFR`<:aKO
	an5L'f'H$C6<N=ieH_Ns:8r/-F>E>*(o2.ET1%BpgNq9:F_:'kIT8!)),C3GlcmYHU;j&0aK<:WMel
	3Ze-Om%!i`\)#_:DlKDdZeWg@P_C[O5M'iUB8>_-2HE;O1>jM;$8=V+SNHPXHb;/UA2:]pC+19D\Bd
	qtPsUXEa5qKga#B2SZ=.SRp%A!RdqL(!Ea7bt3ql@F;Uj=M08??L+[=OW75>HDQ5.GO:69br-505tC
	KaJC@jCMg**JRJQascSIH0iWNK&pUTkRi35A[F<CY2`i#$#_f\:6RcLI9_:V"&,4[]HN!OfQlP*Qtb
	VGkuA?F_!6F<)ocgU#D#V(':e98CFiHpC[p2PjB,D#;1";i6n,<g%nVGOf<>lBjHXs8BA(fD;DdgE&
	_+N0,P*1$jPnaB'W+c:JFR2daE`AbT2\$"\Atc30k=3]2,.R>E*Nf"c^j@=F?-EhKuZgDlI9;bg[c+
	#@$`/2$kB@MhsTDp>G]B0oZ#`%fgeAhHcB?i5($LInM,P9MS?gJ,auQDYK=UG;=]pX[t0`Ddn+h)@5
	*f*'#BK+[7+e+25ad[WOTm!ShS%o^pU.1GgqQgF=j%Y7Q;ZTcp"RlDZ'E)=kG_qYSr#Wb]AY@*-(nJ
	s#A#q"YFMfsgRh]Kt<,gI6=12lg#`?;TLp_*KtthTY]\gR%W:!I8"sroVaPFBo;RFjF5*7K$9'/M"a
	&=8b8<:r],W]a;6"MF%U=g2!m3l-^kmCMU]X>\[CWf<8QErql0Iq<+A?>3\a$'Y&&M_Vt+amGf@imo
	bQW3d]\44-GFK,M+U?MapjQn$rZ2MO]A366&(1N))Y5K\G-`9XF4l;%/:LK#,8h*dQfYG3mlOSIcZr
	>bLn*pC6ZP>8Kg\bG_@CKr2I%"\bg,9T7gnab8=Dm!=-WrDiH(kdbct`7o4FNN0!X<E0c>,]?+,R:L
	D]PZ0,$>?`&92Jr:c#b4?2.q.C==rdOGT-GEe&/PNlY926R6do>OS3)p0O\R3Y8&I`m-S_GCD(TPUp
	#Oh0$f?=CMoA(S7gE=*>mMVCq0YVL.YYcK,Y=`@\X=h:G[ec2^3h=YF;C/g`2VT*c?qXpKj]]U?cVq
	fC?DNujYDHgQ#<&Y/Z3l"'p#PjIED$6^p\*"iPGcpqYD$G(gq`=S8``85)IVK,7k,W='%mZ6N^%W9s
	ZH4Ib4'RZ77*WgpE7az8OZBBY!QNJ
	ASCII85End
end